#!/usr/bin/python
# -*- coding: utf-8 -*-

# signalr_aio/transports/_parameters.py
# Stanislav Lazarov


from json import dumps
from urllib.parse import urlparse, urlunparse, urlencode
import requests


class WebSocketParameters:
    def __init__(self, connection):
        self.protocol_version = '1.5'
        self.raw_url = self._clean_url(connection.url)
        self.conn_data = self._get_conn_data(connection.hub)
        self.session = connection.session
        self.headers = None
        self.socket_conf = None
        self._negotiate()
        self.socket_url = self._get_socket_url()

    @staticmethod
    def _clean_url(url):
        if url[-1] == '/':
            return url[:-1]
        else:
            return url

    @staticmethod
    def _get_conn_data(hub):
        conn_data = dumps([{'name': hub}])
        return conn_data

    @staticmethod
    def _format_url(url, action, query):
        return '{url}/{action}?{query}'.format(url=url, action=action, query=query)

    def _negotiate(self):
        if self.session is None:
            self.session = requests.Session()
        query = urlencode({
            'connectionData': self.conn_data,
            'clientProtocol': self.protocol_version,
        })
        url = self._format_url(self.raw_url, 'negotiate', query)
        self.headers = dict(self.session.headers)
        request = self.session.get(url)
        self.headers['Cookie'] = self._get_cookie_str(request.cookies)
        self.socket_conf = request.json()

    @staticmethod
    def _get_cookie_str(request):
        return '; '.join([
            '%s=%s' % (name, value)
            for name, value in request.items()
        ])

    def _get_socket_url(self):
        ws_url = self._get_ws_url_from()
        query = urlencode({
            'transport': 'webSockets',
            'connectionToken': self.socket_conf['ConnectionToken'],
            'connectionData': self.conn_data,
            'clientProtocol': self.socket_conf['ProtocolVersion'],
        })

        return self._format_url(ws_url, 'connect', query)

    def _get_ws_url_from(self):
        parsed = urlparse(self.raw_url)
        scheme = 'wss' if parsed.scheme == 'https' else 'ws'
        url_data = (scheme, parsed.netloc, parsed.path, parsed.params, parsed.query, parsed.fragment)

        return urlunparse(url_data)



# Standard Library Imports
import base64
import collections
import datetime
import json
import zlib
from urllib.parse import urljoin
from typing import List, Dict, Union
from jellyfish import jaro_similarity, jaro_winkler_similarity
import re
from string import punctuation
import numpy as np
import pandas as pd

# Internal Project Imports
from .constants import *
from .logger import logger
from .exceptions import LiveF1Error
from ..adapters import LivetimingF1adapters

def build_session_endpoint(session_path):
    """
    Constructs a full endpoint URL for accessing session data.

    Parameters
    ----------
    session_path : str
        The path for the specific session data.

    Returns
    -------
    str
        The complete URL for the session endpoint.
    """
    return urljoin(urljoin(BASE_URL, STATIC_ENDPOINT), session_path)


def json_parser_for_objects(data: Dict) -> Dict:
    """
    Converts the keys of a dictionary to lowercase.

    Parameters
    ----------
    data : Dict
        The original dictionary with keys.

    Returns
    -------
    Dict
        A new dictionary with all keys converted to lowercase.
    """
    return {key.lower(): value for key, value in data.items()}


def get_data(path, stream):
    """
    Fetches data from a specified endpoint.

    Parameters
    ----------
    path : str
        The endpoint to retrieve data from.
    stream : bool
        Indicates whether to return a stream of records or a single response.

    Returns
    -------
    Union[dict, str]
        A dictionary of records if `stream` is True, else a string response.
    """
    adapters = LivetimingF1adapters()
    endpoint = path
    res_text = adapters.get(endpoint=endpoint)

    if stream:
        records = res_text.split('\r\n')[:-1]  # Split response into lines, ignoring the last empty line.
        tl = 12  # Length of the key in the response.
        return dict((r[:tl], r[tl:]) for r in records)
    else:
        return res_text  # Return the full response text if not streaming.


def get_car_data_stream(path):
    """
    Fetches car data from a specified endpoint and returns it as a dictionary.

    Parameters
    ----------
    path : str
        The endpoint to retrieve car data from.

    Returns
    -------
    dict
        A dictionary where keys are the first 12 characters of each record and values are the remaining data.
    """
    adapters = LivetimingF1adapters()
    endpoint = path
    res_text = adapters.get(endpoint=endpoint)
    records = res_text.split('\r\n')[:-1]  # Split response into lines, ignoring the last empty line.

    tl = 12  # Length of the key in the response.
    return dict((r[:tl], r[12:]) for r in records)


def parse(text: str, zipped: bool = False) -> Union[str, dict]:
    """
    Parses a given text input and decompresses it if necessary.

    Parameters
    ----------
    text : str
        The input text to be parsed.
    zipped : bool, optional
        Indicates if the input is a zipped string, by default False.

    Returns
    -------
    Union[str, dict]
        The parsed output as a dictionary if input is JSON, otherwise as a string.
    """
    if text[0] == '{':  # Check if the text is in JSON format.
        return json.loads(text)  # Return parsed JSON as a dictionary.
    if text[0] == '"':  # Check if the text is a quoted string.
        text = text.strip('"')  # Remove surrounding quotes.
    if zipped:
        # Decompress the zipped base64 string and parse it.
        text = zlib.decompress(base64.b64decode(text), -zlib.MAX_WBITS)
        return parse(text.decode('utf-8-sig'))
    return text  # Return the text as is if it's not zipped.


def parse_hash(hash_code):
    """
    Parses a hashed string and decompresses it.

    Parameters
    ----------
    hash_code : str
        The hash string to be parsed.

    Returns
    -------
    dict
        The decompressed and parsed data as a dictionary.
    """
    tl = 12  # Length of the key in the response.
    return parse(hash_code, zipped=True)


def parse_helper_for_nested_dict(info, record, prefix=""):
    """
    Recursively parses a nested dictionary and flattens it into a single-level dictionary.

    Parameters
    ----------
    info : dict
        The nested dictionary to parse.
    record : dict
        The record to which parsed information will be added.
    prefix : str, optional
        A prefix for keys in the flattened dictionary, by default "".

    Returns
    -------
    dict
        The updated record with flattened keys from the nested dictionary.
    """
    for info_k, info_v in info.items():
        if isinstance(info_v, list):
            # Flatten list entries into the record with incremental suffixes.
            record = {**record, **{**{info_k + "_" + str(sector_no + 1) + "_" + k: v 
                                      for sector_no in range(len(info_v)) 
                                      for k, v in info_v[sector_no].items()}}}
        elif isinstance(info_v, dict):
            # Recursively parse nested dictionaries.
            record = parse_helper_for_nested_dict(info_v, record, prefix=prefix + info_k + "_")
        else:
            record = {**record, **{prefix + info_k: info_v}}  # Add scalar values to the record.
    return record

def identifer_text_format(text):
    """
    Formats text for comparison by splitting into words and removing stopwords.

    Parameters
    ----------
    text : str
        The input text to format.

    Returns
    -------
    list
        A list of words from the input text with stopwords removed.
    """
    querywords = re.split(rf'[\s{punctuation}]+', text.casefold())
    return [word for word in querywords if word not in QUERY_STOPWORDS]

def find_most_similar_vectorized(df, target):
    """
    Find the most similar string in a Pandas DataFrame using Jaccard and Jaro-Winkler similarity.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to search in.
    target : str
        The string to search for.

    Returns
    -------
    dict
        A dictionary containing:
            - "isFound" (int): 1 if a match is found, 0 otherwise.
            - "how" (str): The method used for matching ("jaccard" or "jaro").
            - "value" (str): The most similar value found.
            - "similarity" (float): The similarity score of the match.
            - "row" (int): The row index of the match.
            - "column" (str): The column name of the match.

    Raises
    ------
    livef1Exception
        If no match is found and suggestions are provided.
    """

    def jaccard_similarity(cell):
        """
        Calculates the Jaccard similarity between two sets of words.

        Parameters
        ----------
        cell : str
            The text to compare against the target.

        Returns
        -------
        float
            The Jaccard similarity score.
        """
        intersection_cardinality = len(
            set.intersection(
                *[
                    set(identifer_text_format(target)),
                    set(identifer_text_format(cell))
                ]
            )
        )
        union_cardinality = len(
            set.union(
                *[
                    set(identifer_text_format(target)),
                    set(identifer_text_format(cell))
                ]
            )
        )
        return intersection_cardinality/float(union_cardinality)

    def jarow_similarity(cell):
        """
        Calculates the Jaro-Winkler similarity between two strings.

        Parameters
        ----------
        cell : str
            The text to compare against the target.

        Returns
        -------
        float
            The Jaro-Winkler similarity score.
        """
        return jaro_winkler_similarity(
            " ".join(identifer_text_format(target)),
            " ".join(identifer_text_format(cell))
            )

    def argmax_n(arr: np.array, n: int, axis=None):
        """
        Finds the indices of the top-n maximum values in an array.

        Parameters
        ----------
        arr : np.array
            The array to search in.
        n : int
            The number of maximum values to find.
        axis : int, optional
            The axis to search along, by default None.

        Returns
        -------
        list
            A list of indices corresponding to the top-n maximum values.
        """
        argmaxes = []
        for _ in range(n):
            row, col = divmod(arr.argmax(), arr.shape[1])
            argmaxes.append(row)
            # arr = np.delete(arr, row, axis=0)
            arr[row,:] = 0
            # print(row, col)
        return argmaxes

    logger.debug(f"Searching of identifier '{target}' has started.")

    similarity_df = df.map(jaccard_similarity)
    jaccard_score = similarity_df.max().max()
    row, col = divmod(similarity_df.values.argmax(), similarity_df.shape[1])
    most_similar = df.iloc[row, col]


    if jaccard_score:
        return {
            "isFound": 1,
            "how" : "jaccard",
            "value": most_similar,
            "similarity": jaccard_score,
            "row": df.iloc[row].name,
            "column": df.columns[col]
        }
    else:
        logger.info("The identifier couldn't be found.")

        jaro_df = df.map(jarow_similarity)
        jaro_score = jaro_df.max().max()

        if jaro_score >= 0.9:
            row, col = divmod(jaro_df.values.argmax(), jaro_df.shape[1])
            most_similar = df.iloc[row, col]
            logger.info(f"The identifier is very close to '{most_similar}' at column '{(df.columns[col]).upper()}'")

            return {
                "isFound": 1,
                "how" : "jaro",
                "value": most_similar,
                "similarity": jaro_score,
                "row": row,
                "column": df.columns[col]
            }

        else:
            poss_args = argmax_n(jaro_df.values, 3, axis=1)
            possible_df = df.iloc[poss_args]

            err_text = f"\nThe searched query '{target}' not found in the table. Did you mean one of these :\n\n"
            for idx, prow in possible_df.iterrows():
                for col in possible_df.columns:
                    err_text += f"\t{col} : {prow[col]}\n"
                # err_text += f"\t> Suggested search queries : {identifer_text_format(prow.meeting_name) + identifer_text_format(prow.meeting_circuit_shortname)}\n\n"
                err_text += f"\t> Suggested search queries : {[identifer_text_format(prow[col])for col in possible_df.columns if not col in EXCLUDED_COLUMNS_FOR_SEARCH_SUGGESTION]}\n\n"
            raise LiveF1Error(err_text)

            return {
                "isFound": 0,
                "how": None,
                "value": None,
                "similarity": None,
                "row": None,
                "column": None
            }

def print_found_model(df, key, cols):
    found_meeting_info = df.loc[[key], cols].drop_duplicates().iloc[0]
    found_info = "\n".join([f"\t{col} : {found_meeting_info[col]}" for col in cols])
    logger.info(f"""Selected meeting/session is:\n{found_info}""")


def to_datetime(var):
    if isinstance(var, pd.Series):
        return pd.to_datetime(var.values, format='ISO8601').tz_localize(None).round("ms")
    elif isinstance(var, np.ndarray):
        return pd.to_datetime(var, format='ISO8601').tz_localize(None).round("ms")
        


import logging
import os

# Logger setup
# LOG_LEVEL = logging.DEBUG if os.getenv("DEBUG", "False").lower() == "true" else logging.INFO
LOG_LEVEL = logging.INFO
STREAM_LOG_FORMAT = "%(asctime)s - %(message)s"
FILE_LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

# Configure logger
logger = logging.getLogger("livef1")
logger.setLevel(LOG_LEVEL)

# Handlers
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter(STREAM_LOG_FORMAT,"%H:%M:%S"))

file_handler = logging.FileHandler("livef1.log")
file_handler.setFormatter(logging.Formatter(FILE_LOG_FORMAT,"%Y-%m-%d %H:%M:%S"))

# Add handlers to logger
logger.addHandler(console_handler)
logger.addHandler(file_handler)

def set_log_level(level):
    """
    Set the logging level for the livef1 logger.
    
    Parameters
    ----------
    level : Union[str, int]
        The logging level to set. Can be either a string ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        or the corresponding integer value.
    
    Examples
    --------
    >>> set_log_level('DEBUG')  # Set to debug level
    >>> set_log_level(logging.INFO)  # Set to info level
    """
    if isinstance(level, str):
        level = level.upper()
        numeric_level = getattr(logging, level, None)
        if not isinstance(numeric_level, int):
            raise ValueError(f'Invalid log level: {level}')
        level = numeric_level
        
    logger.setLevel(level)


import pandas as pd
import requests
from bs4 import BeautifulSoup
from datetime import datetime
from string import digits


def _parse_tables_from_wikipedia(url):
    
    response = requests.get(url)
    soup = BeautifulSoup(response.content, 'html.parser')

    # Store parsed tables with associated titles
    tables_dict = {}
    all_tables = soup.find_all('table', class_='wikitable')

    for i, table in enumerate(all_tables):
        # Try caption first
        caption = table.find('caption')
        if caption:
            title = caption.get_text(strip=True)
        else:
            # No caption: find closest previous header (h2-h6)
            prev = table.find_previous(['h2', 'h3', 'h4', 'h5', 'h6'])
            if prev:
                title = prev.get_text(strip=True).replace('[edit]', '')
            else:
                title = f'Table {i+1}'  # fallback generic title

        # Read the table with pandas
        df = pd.read_html(str(table))[0]
        tables_dict[title] = df
    
    return tables_dict

def get_table_from_wikipedia(url, table_name):
    tables_dict = _parse_tables_from_wikipedia(url)
    
    try:
        return tables_dict.get(table_name)
    except:
        return None

def parse_schedule_from_f1com(season: int) -> pd.DataFrame:
    """Parse the schedule for the given season from the official Formula 1 website.
    Args:
        season (int): The year of the season to parse.
    Returns:
        pd.DataFrame: A DataFrame containing the schedule with columns:
            - "Meeting Circuit Shortname"
            - "Meeting Offname"
            - "Session Name"
            - "Start Date"
            - "End Date"
    """

    BASE_URL = 'https://www.formula1.com'
    SEASON_URL = BASE_URL + f'/en/racing/{season}'

    """Given a race weekend URL, return a dict of session names → datetime."""
    resp = requests.get(SEASON_URL)
    soup = BeautifulSoup(resp.content, 'html.parser')

    meetings = []
    for meeting_obj in soup.find_all('a', class_='group'):

        # short_name = meeting_obj.find(class_="typography-module_display-xl-bold__Gyl5W").text
        offname = meeting_obj.find(class_="typography-module_body-xs-semibold__Fyfwn").text
        meeting_url = BASE_URL + meeting_obj["href"]

        sub_resp = requests.get(meeting_url)
        sub_soup = BeautifulSoup(sub_resp.content, 'html.parser')
        short_name = sub_soup.title.text.split(" - ")[0].split("Grand Prix")[0].strip()


        ul = sub_soup.find('ul', class_="contents")
        items = [li for li in ul.find_all('li')]
        sessions = {}

        for item in items:
            session_name = item.find(class_="typography-module_display-m-bold__qgZFB").text
            times = item.find_all("time")

            day = item.find(class_="typography-module_technical-l-bold__AKrZb").text
            month = item.find(class_="typography-module_technical-s-regular__6LvKq").text
            season = 2025

            if len(times) > 1:
                start = times[0].text
                end = times[1].text
                start_date = datetime.strptime(f"{day} {month} {season} - {start}", "%d %b %Y - %H:%M")
                end_date = datetime.strptime(f"{day} {month} {season} - {end}", "%d %b %Y - %H:%M")

            else:
                start = times[0].text
                end = None
                start_date = datetime.strptime(f"{day} {month} {season} - {start}", "%d %b %Y - %H:%M")
                end_date = None
            
            sessions[session_name] = {
                "start_date" : start_date,
                "end_date" : end_date
            }
        

        meetings.append(
            {
                "short_name" : short_name,
                "offname" : offname,
                "sessions" : sessions
            }
        )
        

    row = []
    for meeting in meetings:

        meeting_shortname = meeting.get("short_name")
        meeting_offname = meeting.get("offname")
        sessions = meeting.get("sessions")
        
        for session_name, dates in sessions.items():
            row.append(
                [
                    meeting_shortname,
                    meeting_offname,
                    session_name,
                    dates["start_date"],
                    dates["end_date"]
                    ]
            )

    schedule_df = pd.DataFrame(row, columns=["Meeting Shortname", "Meeting Offname", "Session Name", "Start Date", "End Date"])
    return schedule_df


# Standard Library Imports
import dateutil
import sys
import json

# Third-Party Library Imports
import pandas as pd
from typing import List, Dict

# Internal Project Imports
from ..adapters import download_data
from ..models.session import Session
from ..models.circuit import Circuit
from ..utils.helper import json_parser_for_objects, build_session_endpoint
from ..utils.constants import SESSIONS_COLUMN_MAP


class Meeting:
    """
    Represents a meeting in a specific season with relevant details and associated sessions.

    Attributes
    ----------
    season : :class:`~Season`
        The season this meeting belongs to.
    year : :class:`int`
        The year of the meeting.
    code : :class:`int`
        The unique code for the meeting.
    key : :class:`str`
        The unique identifier for the meeting.
    number : :class:`int`
        The sequential number of the meeting in the season.
    location : :class:`str`
        The location (e.g., circuit name) of the meeting.
    officialname : :class:`str`
        The official name of the meeting.
    name : :class:`str`
        The name of the meeting.
    country : :class:`dict`
        Details about the country where the meeting takes place (e.g., key, code, name).
    circuit : :class:`dict`
        Details about the circuit where the meeting takes place (e.g., key, short name).
    sessions : :class:`list`
        List of session objects associated with the meeting.
    loaded : :class:`bool`
        Indicates whether the meeting data has been loaded.
    """

    def __init__(
        self,
        season: "Season" = None,
        year: int = None,
        code: int = None,
        key: str = None,
        number: int = None,
        location: str = None,
        officialname: str = None,
        name: str = None,
        country: Dict = None,
        circuit: Dict = None,
        sessions: List = None,
        loaded: bool = False,
        **kwargs  # In case new information comes from the API in future
    ):
        self.season = season
        self.loaded = loaded

        # Iterate over the kwargs and set them as attributes of the instance
        for key, value in locals().items():
            if value:
                setattr(self, key.lower(), value)

        # Load Circuit
        self.circuit = Circuit(self.circuit["Key"],self.circuit["ShortName"])
        self.circuit._load_start_coordinates()

        if hasattr(self, "sessions"):
            self.sessions_json = self.sessions
            self.sessions = []
            self.set_sessions()

        self.parse_sessions()

    def load(self, force=False):
        """
        Load or reload meeting data from the API.

        .. note::
            Reloading is useful when updated data is required.

        Parameters
        ----------
        force : bool, optional
            If True, forces the reload of meeting data even if already loaded. Defaults to False.
        
        
        """
        if (not self.loaded) | (force):
            if force:
                print("Force load...")

            if hasattr(self, "year"):
                self.json_data = download_data(self.year, self.location)
            elif hasattr(self, "season"):
                self.json_data = download_data(self.season.year, self.location)

            for key, value in json_parser_for_objects(self.json_data).items():
                setattr(self, key.lower(), value)

            self.sessions_json = self.sessions
            self.sessions = []

            self.parse_sessions()
            self.set_sessions()
        else:
            print("The meeting has already been loaded. If you want to load anyway, use `force=True`.")

    def set_sessions(self):
        """
        Create session objects for the meeting using the session JSON data.

        .. note::
            This method populates the `sessions` attribute with `Session` objects derived from `sessions_json`.
        """
        for session_data in self.sessions_json:
            self.sessions.append(
                Session(
                    season=self.season,
                    meeting=self,
                    **json_parser_for_objects(session_data)
                )
            )

    def parse_sessions(self):
        """
        Parse session data to generate a detailed DataFrame of session metadata.

        .. note::
            The resulting DataFrame is stored in the `sessions_table` attribute and indexed by season year, meeting location, and session type.
        """
        session_all_data = []

        for session in self.sessions_json:
            session_data = {
                "season_year": dateutil.parser.parse(session["StartDate"]).year,
                "meeting_code": self.code,
                "meeting_key": self.key,
                "meeting_number": self.number,
                "meeting_location": self.location,
                "meeting_offname": self.officialname,
                "meeting_name": self.name,
                "meeting_country_key": self.country["Key"],
                "meeting_country_code": self.country["Code"],
                "meeting_country_name": self.country["Name"],
                "circuit": self.circuit,
                "session_key": session.get("Key", None),
                "session_type": session["Type"] + " " + str(session["Number"]) if "Number" in session else session["Type"],
                "session_name": session.get("Name", None),
                "session_startDate": session.get("StartDate", None),
                "session_endDate": session.get("EndDate", None),
                "gmtoffset": session.get("GmtOffset", None),
                "path": session.get("Path", None),
            }
            session_all_data.append(session_data)

        self.meeting_table = pd.DataFrame(session_all_data).set_index(["season_year", "meeting_location", "session_type"])
        self.meeting_table["session_startDate"] = pd.to_datetime(self.meeting_table["session_startDate"])
        self.meeting_table["session_endDate"] = pd.to_datetime(self.meeting_table["session_endDate"])

        self.sessions_table = self.meeting_table[["meeting_key","session_key","session_name","session_startDate","session_endDate","gmtoffset","path"]].set_index("session_key")
        # self.meeting_table = self.meeting_table.reset_index().rename(columns = SESSIONS_COLUMN_MAP)



        # self.meeting_table = 

    def __repr__(self):
        """
        Return a detailed string representation of the meeting.

        Returns
        -------
        str
            The string representation of the meeting's session table.
        """
        if "IPython" not in sys.modules:
            # definitely not in IPython
            return self.meeting_table.__str__() # Print the meetings table.
        else:
            display(self.meeting_table) # Display the meetings table.
            return ""

    def __str__(self):
        """
        Return a readable string representation of the meeting.

        Returns
        -------
        str
            The string representation of the meeting's session table.
        """
        return self.meeting_table.__str__()


# Default API URL and endpoints
BASE_URL = "https://livetiming.formula1.com"
STATIC_ENDPOINT = "/static/"
SIGNALR_ENDPOINT = "/signalr/"

# Circuit Start Coordinates URL
START_COORDINATES_URL = "https://raw.githubusercontent.com/GoktugOcal/LiveF1/refs/heads/main/livef1/data/starting_coordinates.json"

DEFAULT_METHOD = "livef1"

REALTIME_CALLBACK_DEFAULT_PARAMETERS = [
  # "topic_name",
  # "data",
  # "timestamp"
  "records"
]

QUERY_STOPWORDS = [
    "formula",
    "1",
    "grand",
    "prix"
]

SESSIONS_COLUMN_MAP = {
    'season_year': 'Season Year',
    'meeting_code': 'Meeting Code',
    'meeting_key': 'Meeting Key',
    'meeting_number': 'Meeting Number',
    'meeting_location': 'Meeting Location',
    'meeting_offname': 'Meeting Offname',
    'meeting_name': 'Meeting Name',
    'meeting_country_key': 'Meeting Country Key',
    'meeting_country_code': 'Meeting Country Code',
    'meeting_country_name': 'Meeting Country Name',
    'meeting_circuit_key': 'Meeting Circuit Key',
    'meeting_circuit_shortname': 'Meeting Circuit Shortname',
    'session_key': 'Session Key',
    'session_type': 'Session Type',
    'session_name': 'Session Name',
    'session_startDate': 'Session Startdate',
    'session_endDate': 'Session Enddate',
    'gmtoffset': 'Gmtoffset',
    'path': 'Path'
}

EXCLUDED_COLUMNS_FOR_SEARCH_SUGGESTION = ["meeting_offname"]

session_index = {
  "Feeds": {
    "SessionInfo": {
      "KeyFramePath": "SessionInfo.json",
      "StreamPath": "SessionInfo.jsonStream"
    },
    "ArchiveStatus": {
      "KeyFramePath": "ArchiveStatus.json",
      "StreamPath": "ArchiveStatus.jsonStream"
    },
    "TrackStatus": {
      "KeyFramePath": "TrackStatus.json",
      "StreamPath": "TrackStatus.jsonStream"
    },
    "SessionData": {
      "KeyFramePath": "SessionData.json",
      "StreamPath": "SessionData.jsonStream"
    },
    "ContentStreams": {
      "KeyFramePath": "ContentStreams.json",
      "StreamPath": "ContentStreams.jsonStream"
    },
    "AudioStreams": {
      "KeyFramePath": "AudioStreams.json",
      "StreamPath": "AudioStreams.jsonStream"
    },
    "ExtrapolatedClock": {
      "KeyFramePath": "ExtrapolatedClock.json",
      "StreamPath": "ExtrapolatedClock.jsonStream"
    },
    "TyreStintSeries": {
      "KeyFramePath": "TyreStintSeries.json",
      "StreamPath": "TyreStintSeries.jsonStream"
    },
    "SessionStatus": {
      "KeyFramePath": "SessionStatus.json",
      "StreamPath": "SessionStatus.jsonStream"
    },
    "TimingDataF1": {
      "KeyFramePath": "TimingDataF1.json",
      "StreamPath": "TimingDataF1.jsonStream"
    },
    "TimingData": {
      "KeyFramePath": "TimingData.json",
      "StreamPath": "TimingData.jsonStream"
    },
    "DriverList": {
      "KeyFramePath": "DriverList.json",
      "StreamPath": "DriverList.jsonStream"
    },
    "LapSeries": {
      "KeyFramePath": "LapSeries.json",
      "StreamPath": "LapSeries.jsonStream"
    },
    "TopThree": {
      "KeyFramePath": "TopThree.json",
      "StreamPath": "TopThree.jsonStream"
    },
    "TimingAppData": {
      "KeyFramePath": "TimingAppData.json",
      "StreamPath": "TimingAppData.jsonStream"
    },
    "TimingStats": {
      "KeyFramePath": "TimingStats.json",
      "StreamPath": "TimingStats.jsonStream"
    },
    "Heartbeat": {
      "KeyFramePath": "Heartbeat.json",
      "StreamPath": "Heartbeat.jsonStream"
    },
    "WeatherData": {
      "KeyFramePath": "WeatherData.json",
      "StreamPath": "WeatherData.jsonStream"
    },
    "WeatherDataSeries": {
      "KeyFramePath": "WeatherDataSeries.json",
      "StreamPath": "WeatherDataSeries.jsonStream"
    },
    "Position.z": {
      "KeyFramePath": "Position.z.json",
      "StreamPath": "Position.z.jsonStream"
    },
    "CarData.z": {
      "KeyFramePath": "CarData.z.json",
      "StreamPath": "CarData.z.jsonStream"
    },
    "TlaRcm": {
      "KeyFramePath": "TlaRcm.json",
      "StreamPath": "TlaRcm.jsonStream"
    },
    "RaceControlMessages": {
      "KeyFramePath": "RaceControlMessages.json",
      "StreamPath": "RaceControlMessages.jsonStream"
    },
    "PitLaneTimeCollection": {
      "KeyFramePath": "PitLaneTimeCollection.json",
      "StreamPath": "PitLaneTimeCollection.jsonStream"
    },
    "CurrentTyres": {
      "KeyFramePath": "CurrentTyres.json",
      "StreamPath": "CurrentTyres.jsonStream"
    },
    "TeamRadio": {
      "KeyFramePath": "TeamRadio.json",
      "StreamPath": "TeamRadio.jsonStream"
    }
  }
}

TOPICS_MAP = {
  "SessionInfo": {
    "key": "Session_Info",
    "description": "Details about the current session.",
    "default_is_stream": True
  },
  "ArchiveStatus": {
    "key": "Archive_Status",
    "description": "Status of archived session data.",
    "default_is_stream": True
  },
  "TrackStatus": {
    "key": "Track_Status",
    "description": "Current conditions and status of the track.",
    "default_is_stream": True
  },
  "SessionData": {
    "key": "Session_Data",
    "description": "Raw data for the ongoing session.",
    "default_is_stream": True
  },
  "ContentStreams": {
    "key": "Content_Streams",
    "description": "Streams of multimedia content.",
    "default_is_stream": True
  },
  "AudioStreams": {
    "key": "Audio_Streams",
    "description": "Live audio broadcast streams.",
    "default_is_stream": True
  },
  "ExtrapolatedClock": {
    "key": "Extrapolated_Clock",
    "description": "Predicted session time data.",
    "default_is_stream": True
  },
  "TyreStintSeries": {
    "key": "Tyre_Stints",
    "description": "Data on tyre usage over stints.",
    "default_is_stream": True
  },
  "SessionStatus": {
    "key": "Session_Status",
    "description": "Live status of the session.",
    "default_is_stream": True
  },
  "TimingDataF1": {
    "key": "Timing_Data_F1",
    "description": "Timing information specific to Formula 1.",
    "default_is_stream": True
  },
  "TimingData": {
    "key": "Timing_Data",
    "description": "General timing data for the session.",
    "default_is_stream": True
  },
  "DriverList": {
    "key": "Driver_List",
    "description": "List of active drivers in the session.",
    "default_is_stream": False
  },
  "LapSeries": {
    "key": "Lap_Series",
    "description": "Data series for laps completed.",
    "default_is_stream": True
  },
  "TopThree": {
    "key": "Top_Three",
    "description": "Information about the top three drivers.",
    "default_is_stream": True
  },
  "TimingAppData": {
    "key": "Timing_App",
    "description": "Timing data from the application.",
    "default_is_stream": True
  },
  "TimingStats": {
    "key": "Timing_Stats",
    "description": "Statistical analysis of timing data.",
    "default_is_stream": True
  },
  "Heartbeat": {
    "key": "Heartbeat",
    "description": "Regular status signal of the system.",
    "default_is_stream": True
  },
  "WeatherData": {
    "key": "Weather_Data",
    "description": "Current weather information.",
    "default_is_stream": True
  },
  "WeatherDataSeries": {
    "key": "Weather_History",
    "description": "Historical weather data series.",
    "default_is_stream": True
  },
  "Position.z": {
    "key": "Position",
    "description": "Position data of cars.",
    "default_is_stream": True
  },
  "CarData.z": {
    "key": "Car_Data",
    "description": "Car sensor data.",
    "default_is_stream": True
  },
  "TlaRcm": {
    "key": "Team_Audio_RCM",
    "description": "Team live audio and race control messages.",
    "default_is_stream": True
  },
  "RaceControlMessages": {
    "key": "Race_Control",
    "description": "Messages from race control.",
    "default_is_stream": True
  },
  "PitLaneTimeCollection": {
    "key": "Pit_Lane_Times",
    "description": "Timing data for pit lane activity.",
    "default_is_stream": True
  },
  "CurrentTyres": {
    "key": "Current_Tyres",
    "description": "Details of the tyres currently in use.",
    "default_is_stream": True
  },
  "DriverRaceInfo": {
    "key": "Driver_Race_Info",
    "description": "Information about individual driver performance.",
    "default_is_stream": True
  },
  "TeamRadio": {
    "key": "Team_Radio",
    "description": "Radio communications with the team.",
    "default_is_stream": True
  },
  "ChampionshipPrediction": {
    "key": "Championship_Prediction",
    "description": "Predictions for championship outcomes.",
    "default_is_stream": True
  },
  "OvertakeSeries": {
    "key": "Overtake_Series",
    "description": "Data series tracking overtakes.",
    "default_is_stream": True
  },
  "DriverScore": {
    "key": "Driver_Score",
    "description": "Scores reflecting driver performance.",
    "default_is_stream": True
  },
  "SPFeed": {
    "key": "SP_Feed",
    "description": "Special data feed for the session.",
    "default_is_stream": True
  },
  "PitStopSeries": {
    "key": "Pit_Stop_Series",
    "description": "Data series for multiple pit stops.",
    "default_is_stream": True
  },
  "PitStop": {
    "key": "Pit_Stop",
    "description": "Details about individual pit stops.",
    "default_is_stream": True
  },
  "LapCount": {
    "key": "Lap_Count",
    "description": "Number of laps completed in the session.",
    "default_is_stream": True
  },
  "DriverTracker": {
    "key": "Driver_Tracker",
    "description": "Real-time tracking of driver positions.",
    "default_is_stream": True
  }
}

channel_name_map = {
  '0': 'rpm',
  '2': 'speed',
  '3': 'n_gear',
  '4': 'throttle',
  '5': 'brake',
  '45': 'drs'
}

interpolation_map = {
  #Position
  'status': 'ffill',
  'X': 'quadratic',
  'Y': 'quadratic',
  'Z': 'quadratic',
  #Car
  'Speed': 'linear',
  'RPM': 'linear',
  'Throttle': 'linear',
  'Brake': 'ffill',
  'DRS': 'ffill',
  'GearNo': 'ffill',
  #Weather
  "AirTemp" : "linear",
  "Humidity" : "linear",
  "Pressure" : "linear",
  "Rainfall" : "linear",
  "TrackTemp" : "linear",
  "WindDirection" : "linear",
  "WindDirection" : "linear",
  "WindSpeed" : "polynomial",
  # "Distance" : "quadratic"
  }

SILVER_SESSION_TABLES = ["laps", "carTelemetry"]
# SILVER_SESSION_TABLES = ["laps", "car_telemetry", "weather", "timing"]

TABLE_GENERATION_FUNCTIONS = {
    "laps": "generate_laps_table",
    "carTelemetry": "generate_car_telemetry_table"
}

TABLE_REQUIREMENTS = {
    "laps": ["TimingData", "RaceControlMessages", "TyreStintSeries","TrackStatus"],
    "carTelemetry": ["CarData.z", "Position.z", "TyreStintSeries", "laps", "TrackStatus", "TimingData", "track_regions"]
}

column_mapping = {
  'rpm': "RPM",
  'speed' : "Speed",
  'n_gear' : "GearNo",
  'throttle': "Throttle",
  'brake' : "Brake",
  'drs' : "DRS"
}

silver_cartel_col_order = [
  'SessionKey',
  'DriverNo',
  'Utc',
  'timestamp',
  'LapNo',
  'Position',
  'RPM',
  'Speed',
  'GearNo',
  'Throttle',
  'Brake',
  'DRS',
  'X',
  'Y',
  'Z',
  "TrackRegion",
  'Compound',
  'TyreAge',
  'Distance',
  'CarStatus',
  'TrackStatus',
  'tag'
  ]

silver_laps_col_order = [
  'SessionKey',
  'DriverNo',
  'LapNo',
  'LapTime',
  'Position',
  'NoPits',
  'PitIn',
  'PitOut',
  'PitStopDuration',
  'Compound',
  'TyreAge',
  'GapToLeader',
  'IntervalToPositionAhead',
  'Sector1_Time',
  'Sector2_Time',
  'Sector3_Time',
  'LapStartTime',
  'LapStartDate',
  'Speed_I1',
  'Speed_I2',
  'Speed_FL',
  'Speed_ST',
  'TrackStatus',
  'IsDeleted',
  'DeletionMessage',
  'Driver'
]


# Standard Library Imports
import json
from datetime import datetime

# Third-Party Library Imports
import pandas as pd
from ..utils.constants import column_mapping


class BasicResult:
    """
    Encapsulates a basic result dataset, typically in JSON format.

    Parameters
    ----------
    data : :class:`dict`
        The JSON-like data to be encapsulated within the result.

    Attributes
    ----------
    value : :class:`dict`
        The data associated with the result, stored as a dictionary (JSON-like structure).
    """

    def __init__(self, data: dict):
        """
        Initializes the BasicResult instance with the provided data.
        """
        self.value = data
        self.df = pd.DataFrame(data).rename(
            columns = column_mapping
        )

    def __get__(self):
        """
        Retrieves the stored value.

        Returns
        -------
        dict
            The JSON-like data encapsulated within the instance.
        """
        return self.value
    
    def __str__(self):
        """
        Returns a string representation of the stored data as a DataFrame.

        Returns
        -------
        str
            A string representation of the data in tabular format (Pandas DataFrame).
        """
        return self.df.__str__()


# class BronzeResult(BasicResult):
#     """
#     Encapsulates bronze level data, typically raw data.

#     Parameters
#     ----------
#     data : :class:`dict`
#         The raw data to be encapsulated within the result.
#     """

#     def __init__(self, data: dict):
#         """
#         Initializes the BronzeResult instance with the provided data.
#         """
#         super().__init__(data)


# class SilverResult(BasicResult):
#     """
#     Encapsulates silver level data, typically cleaned data.

#     Parameters
#     ----------
#     data : :class:`dict`
#         The cleaned data to be encapsulated within the result.
#     """

#     def __init__(self, data: dict):
#         """
#         Initializes the SilverResult instance with the provided data.
#         """
#         super().__init__(data)


# class GoldResult(BasicResult):
#     """
#     Encapsulates gold level data, typically aggregated data.

#     Parameters
#     ----------
#     data : :class:`dict`
#         The aggregated data to be encapsulated within the result.
#     """

#     def __init__(self, data: dict):
#         """
#         Initializes the GoldResult instance with the provided data.
#         """
#         super().__init__(data)





class Table:
    def __init__(self, table_name, data_lake = None):
        self.data_lake = data_lake
        self.table_name = table_name
        self.table = None
        self.callback = None
        self.df = None
    
    def generate_table(self):
        if self.callback:
            self.df = self.callback(self)
            self.data_lake.update_metadata(
                table_name = self.table_name,
                level = None,
                created_at = datetime.now(),
                generated = True
            )
        return self.df

class BronzeTable(Table):
    def __init__(self, table_name, data, parsed_data, data_lake = None):
        super().__init__(table_name, data_lake)
        self.raw = data
        self.df = pd.DataFrame(parsed_data).rename(
            columns = column_mapping
        )
        if "timestamp" in self.df.columns:
            self.df.timestamp = pd.to_timedelta(self.df.timestamp)

class SilverTable(Table):
    def __init__(self, table_name, sources, source_tables = {"bronze": [], "silver": [], "gold": []}, data_lake = None):
        super().__init__(table_name, data_lake)
        self.sources = sources
        self.source_tables = source_tables
        self.df = None
        self.dependency_tables = []

    def refine_sources(self):
        for source in self.sources:
            level = self.data_lake._identify_table_level(source)
            if level:
                self.source_tables[level].append(source)
                if level in ["silver", "gold"]:
                    dependency_table = self.data_lake.get(level = level, table_name = source)
                    if dependency_table:
                        self.dependency_tables.append(dependency_table)
                    else:
                        raise ValueError(f"Source table '{source}' not found in data lake.")
            else:
                raise ValueError(f"Source table '{source}' not found in data lake.")

class GoldTable(Table):
    def __init__(self, table_name, sources, source_tables = {"bronze": [], "silver": [], "gold": []}, data_lake = None):
        super().__init__(table_name, data_lake)
        self.sources = sources
        self.source_tables = source_tables
        self.df = None
        self.dependency_tables = []


    def refine_sources(self):
        for source in self.sources:
            level = self.data_lake._identify_table_level(source)
            if level:
                self.source_tables[level].append(source)
                if level in ["silver", "gold"]:
                    dependency_table = self.data_lake.get(level = level, table_name = source)
                    if dependency_table:
                        self.dependency_tables.append(dependency_table)
                    else:
                        raise ValueError(f"Source table '{source}' not found in data lake.")
            else:
                raise ValueError(f"Source table '{source}' not found in data lake.")


import pandas as pd
from datetime import datetime

from ..utils.logger import logger
from ..utils.constants import TABLE_GENERATION_FUNCTIONS, TABLE_REQUIREMENTS
from .silver_functions import *
from .data_models import *

# class BronzeLake:
#     def __init__(self, session, great_lake):
#         self.great_lake = great_lake
#         self.lake = {}

#     def put(self, data_name, data):
#         """
#         Store the data in the BronzeLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to store.
#         data : object
#             The data to store.
#         """
        
#         self.lake[data_name] = data

#     def get(self, data_name):
#         """
#         Retrieve the data from the BronzeLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to retrieve.

#         Returns
#         -------
#         object
#             The requested data or None if it does not exist.
#         """
#         if self.has_data(data_name):
#             return self.lake[data_name]
#         else:
#             logger.info(f"Data '{data_name}' is not present in BronzeLake.")
#             return None

#     def has_data(self, data_name):
#         """
#         Check if the data exists in the BronzeLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to check.

#         Returns
#         -------
#         bool
#             True if the data exists, False otherwise.
#         """
#         return data_name in self.lake


# class SilverLake:
#     def __init__(self, great_lake, bronze_lake):
#         self.great_lake = great_lake
#         self.bronze_lake = bronze_lake
#         self.lake = {}

#     def get(self, data_name):
#         """
#         Retrieve the data from the SilverLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to retrieve.

#         Returns
#         -------
#         object
#             The requested data or None if it does not exist.
#         """
#         if self.has_data(data_name):
#             return self.lake[data_name]
#         else:
#             logger.info(f"Data '{data_name}' is not present in SilverLake.")
#             return None

#     def clean_data(self, data):
#         cleaned_data = []
#         for record in data:
#             cleaned_record = record  # Placeholder for actual cleaning logic
#             cleaned_data.append(cleaned_record)
#         return cleaned_data

#     def generate_table(self, table_name):
#         """
#         Generate a table using the corresponding function from silver_functions.py.

#         Parameters
#         ----------
#         table_name : str
#             The name of the table to generate.

#         Returns
#         -------
#         DataFrame
#             The generated table as a pandas DataFrame.
#         """
#         if table_name in TABLE_GENERATION_FUNCTIONS:
#             required_data = TABLE_REQUIREMENTS[table_name]
#             for data_name in required_data:
#                 if not self.bronze_lake.has_data(data_name):
#                     self.great_lake.session.get_data(data_name)
#             function_name = TABLE_GENERATION_FUNCTIONS[table_name]
#             return globals()[function_name](self.bronze_lake)
#         else:
#             raise ValueError(f"No generation function found for table: {table_name}")


# class GoldLake:
#     def __init__(self, great_lake, silver_lake):
#         self.great_lake = great_lake
#         self.silver_lake = silver_lake
#         self.lake = {}

#     def get(self, data_name):
#         """
#         Retrieve the data from the GoldLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to retrieve.

#         Returns
#         -------
#         object
#             The requested data or None if it does not exist.
#         """
#         if self.has_data(data_name):
#             return self.lake[data_name]
#         else:
#             logger.info(f"Data '{data_name}' is not present in GoldLake.")
#             return None

#     def aggregate_data(self, data):
#         aggregated_data = []
#         for record in data:
#             aggregated_record = record  # Placeholder for actual aggregation logic
#             aggregated_data.append(aggregated_record)
#         return aggregated_data


# class DataLake:
#     def __init__(self, session):
#         self.raw = {}
#         self.session = session
#         self.bronze_lake = BronzeLake(session=session, great_lake=self)
#         self.silver_lake = SilverLake(great_lake=self, bronze_lake=self.bronze_lake)
#         self.gold_lake = GoldLake(great_lake=self, silver_lake=self.silver_lake)

#     def load_data(self, level: str, data_name: str):
#         if level == "bronze":
#             return self.bronze_lake.get(data_name)
#         elif level == "silver":
#             return self.silver_lake.get(data_name)
#         elif level == "gold":
#             return self.gold_lake.get(data_name)
#         else:
#             raise ValueError("Invalid level. Must be one of 'bronze', 'silver', or 'gold'.")

#     def put(self, level, data_name, data):
#         """
#         Store the data in the DataLake.

#         Parameters
#         ----------
#         data_name : str
#             The name of the data to store.
#         data : object
#             The data to store.
#         """

#         if level == "bronze":
#             self.raw[data_name] = data.value
#             self.bronze_lake.put(data_name, data.df)
#         elif level == "silver":
#             pass
#         elif level == "gold":
#             pass
#         else:
#             raise ValueError("Invalid level. Must be one of 'bronze', 'silver', or 'gold'.")

#     def get(self, level: str, data_name: str):
#         """
#         Retrieve the data from the DataLake.

#         Parameters
#         ----------
#         level : str
#             The level of the lake ('bronze', 'silver', 'gold').
#         data_name : str
#             The name of the data to retrieve.

#         Returns
#         -------
#         object
#             The requested data or None if it does not exist.
#         """
#         if level == "bronze":
#             return self.bronze_lake.get(data_name)
#         elif level == "silver":
#             return self.silver_lake.get(data_name)
#         elif level == "gold":
#             return self.gold_lake.get(data_name)
#         else:
#             raise ValueError("Invalid level. Must be one of 'bronze', 'silver', or 'gold'.")

class SimpleLake:
    def __init__(self, great_lake):
        self.great_lake = great_lake
        self.lake = {}
        self.lake_type = None
    
    def put(self, table_name, table):
        """
        Store the data in the Lake.

        Parameters
        ----------
        data_name : str
            The name of the data to store.
        bronze_table : object
            The data to store.
        """
        table.data_lake = self.great_lake
        self.lake[table_name] = table
        self.great_lake.update_metadata(
            table_name = table_name,
            level = self.lake_type,
            created_at = datetime.now()
            )

    def get(self, table_name):
        """
        Retrieve the data from the Lake.

        Parameters
        ----------
        data_name : str
            The name of the data to retrieve.

        Returns
        -------
        object
            The requested data or None if it does not exist.
        """
        if self.has_data(table_name):
            return self.lake[table_name]
        else:
            logger.info(f"Table '{table_name}' is not present in BronzeLake.")
            return None

    def has_data(self, table_name):
        """
        Check if the data exists in the Lake.

        Parameters
        ----------
        data_name : str
            The name of the data to check.

        Returns
        -------
        bool
            True if the data exists, False otherwise.
        """
        return table_name in self.lake
            

class BronzeLake(SimpleLake):
    def __init__(self, great_lake):
        super().__init__(great_lake)
        self.lake_type = "bronze"

class SilverLake(SimpleLake):
    def __init__(self, great_lake):
        super().__init__(great_lake)
        self.lake_type = "silver"

class GoldLake(SimpleLake):
    def __init__(self, great_lake):
        super().__init__(great_lake)
        self.lake_type = "gold"


class DataLake:
    def __init__(self, session):
        self.session = session
        self.metadata = {}

        self.bronze = BronzeLake(great_lake=self)
        self.silver = SilverLake(great_lake=self)
        self.gold = GoldLake(great_lake=self)
    
    def update_metadata(
        self,
        table_name,
        level,
        created_at=None,
        generated=False
        ):

        if table_name in self.metadata:
            self.metadata[table_name] = {
                "table_type": level if level else self.metadata[table_name]["table_type"],
                "created_at": created_at if created_at else self.metadata[table_name]["created_at"],
                "generated" : generated if generated else self.metadata[table_name]["generated"]
            }
        
        else:
            self.metadata[table_name] = {
                "table_type": level,
                "created_at": created_at,
                "generated" : generated
            }
    
    def put(self, level, table_name, table):
        """
        Store the data in the DataLake.

        Parameters
        ----------
        data_name : str
            The name of the data to store.
        data : object
            The data to store.
        """

        if level == "bronze":
            self.bronze.put(table_name, table)
        elif level == "silver":
            self.silver.put(table_name, table)
        elif level == "gold":
            self.gold.put(table_name, table)
        else:
            raise ValueError("Invalid level. Must be one of 'bronze', 'silver', or 'gold'.")
    
    def get(self, level: str, table_name: str):
        """
        Retrieve the data from the DataLake.

        Parameters
        ----------
        level : str
            The level of the lake ('bronze', 'silver', 'gold').
        data_name : str
            The name of the data to retrieve.

        Returns
        -------
        object
            The requested data or None if it does not exist.
        """
        if level == "bronze":
            return self.bronze.get(table_name)
        elif level == "silver":
            return self.silver.get(table_name)
        elif level == "gold":
            return self.gold.get(table_name)
        else:
            raise ValueError("Invalid level. Must be one of 'bronze', 'silver', or 'gold'.")
    
    # def create_silver_table(self, source_tables, table_name):
    #     """
    #     Decorator factory that creates a SilverTable instance
        
    #     Args:
    #         source_tables: List of source tables to use (BronzeTable instances in BronzeLake)
    #         table_name: Optional name for the table, defaults to function name if None
    #     """
    #     def decorator(callback_func):
    #         # Create a new SilverTable instance
    #         table = SilverTable(table_name, sources=source_tables)
    #         table.callback = callback_func
            
    #         # # Set the create_table method to run the callback
    #         # def create_table_wrapper():
    #         #     return table.create_table()
            
    #         # # Add the create_table method to the returned object
    #         # callback_func.create_table = create_table_wrapper
    #         # callback_func.table = table
            
    #         print("putting table to silver lake...")
    #         self.put(level="silver", table_name=table_name, table=table)
    #         return callback_func
        
    #     return decorator


    def create_bronze_table(self, table_name, raw_data, parsed_data):
        """
        Decorator factory that creates a BronzeTable instance
        
        Args:
            source_tables: List of source tables to use (BronzeTable instances in BronzeLake)
            table_name: Optional name for the table, defaults to function name if None
        """
        self.put(
            level="bronze",
            table_name=table_name,
            table= BronzeTable(
                table_name=table_name,
                data=raw_data,  
                parsed_data=parsed_data
            )
        )
    
    def _identify_table_level(self, table_name):
        """
        Identify the level of a table
        """
        if table_name in self.metadata:
            return self.metadata[table_name]["table_type"]
        else:
            table_level = self.session._identify_data_level(table_name)
            if table_level:
                return table_level
            else:
                return None
            

    def _check_circular_dependencies(self):
        """
        Check for circular dependencies in the tables across different levels.
        
        This method detects cycles in the dependency graph of tables, which would
        make it impossible to generate tables in a valid order.
        
        Returns
        -------
        bool
            True if no circular dependencies exist, False otherwise.
        
        Raises
        ------
        ValueError
            If circular dependencies are detected, with details about the cycle.
        """
        # Track visited tables and current path during DFS
        visited = set()
        path = set()
        path_list = []
        
        def dfs(table):
            """Depth-first search to detect cycles in the dependency graph."""
            # If we've already fully explored this node, no need to check again
            if table in visited:
                return True
            
            # If we encounter a table already in our current path, we have a cycle
            table_name = table.table_name
            if table_name in path:
                cycle_start = path_list.index(table_name)
                cycle = path_list[cycle_start:] + [table_name]
                raise ValueError(f"Circular dependency detected: {' -> '.join(cycle)}")
            
            # Add current table to path
            path.add(table_name)
            path_list.append(table_name)
            
            # If table has dependency_tables attribute, check those too
            if hasattr(table, 'dependency_tables') and table.dependency_tables:
                for dep_table in table.dependency_tables:
                    # dependency_tables contains actual table objects
                    dep_table_name = dep_table.table_name
                    if not dfs(dep_table):
                        return False
            
            # Remove from current path and mark as visited
            path.remove(table_name)
            path_list.pop()
            visited.add(table_name)
            return True
        
        # Start DFS from all tables in the metadata
        for table_name, info in self.metadata.items():
            if info["table_type"] in ["silver", "gold"] and table_name not in visited:
                table = self.get(info["table_type"], table_name)
                if not dfs(table):
                    return False
        
        return True


    def generate_silver_table(self, table_name):
        """
        Generate a silver table from a bronze tables
        """
        self.silver.lake[table_name].generate_table()









from typing import Dict, Optional
import requests
import pandas as pd

from livef1.utils.constants import START_COORDINATES_URL
from livef1.utils.exceptions import livef1Exception


class Circuit:
    """
    Represents a Formula 1 circuit with its characteristics and metadata.

    Attributes
    ----------
    key : str
        The unique identifier for the circuit.
    short_name : str
        Short name/abbreviation of the circuit.
    name : str, optional
        Full name of the circuit.
    length : float, optional
        Length of the circuit in kilometers.
    laps : int, optional
        Standard number of race laps.
    country : Dict, optional
        Dictionary containing country information.
    location : str, optional
        Geographic location of the circuit.
    coordinates : Dict[str, float], optional
        Latitude and longitude of the circuit.
    """

    def __init__(
        self,
        key: str,
        short_name: str
    ):
        self.key = key
        self.short_name = short_name
    
    def _load_start_coordinates(self):
        """
        Load the start coordinates of the circuit from an external API.
        """
        response = requests.get(START_COORDINATES_URL)
        
        if response.status_code == 200:
            data = response.json()
            try:
                self.start_coordinates = data[self.short_name]["start_coordinates"]
                self.start_direction = data[self.short_name]["start_direction"]
            except:
                pass
        else:
            raise Exception(f"Failed to load start coordinates: {response.status_code}")
        
    def _load_circuit_data(self):
        HEADERS = {'User-Agent': 'LiveF1/user'}
        response = requests.get("https://api.multiviewer.app/api/v1/circuits", headers=HEADERS)
        circuits = response.json()

        circuit_ref = next((v for v in circuits.values() if v.get('name').lower() == self.short_name.lower()), None)
        if circuit_ref == None:
            raise livef1Exception(f"Circut {self.short_name} couldn't be found in circuit api.")

        circuit_key = circuit_ref["circuitKey"]
        circuit_ref_years = circuit_ref["years"]

        response = requests.get(f"https://api.multiviewer.app/api/v1/circuits/{circuit_key}/{circuit_ref_years[0]}/", headers=HEADERS)
        self._raw_circuit_data = response.json()

        for corner in self._raw_circuit_data["corners"]:
            corner.update(corner["trackPosition"])
        df_corners = pd.DataFrame(self._raw_circuit_data["corners"]).rename(
                columns = {
                    "length": "Distance",
                    "x": "X",
                    "y": "Y"
                }
            )
        df_corners.Distance = df_corners.Distance / 10
        df_corners["corner_start"] = df_corners.Distance - 50
        df_corners["corner_end"] = df_corners.Distance + 50

        df_corners["corner_end"] = (df_corners["corner_end"] > df_corners["corner_start"].shift(-1).fillna(1000000)) * df_corners["corner_start"].shift(-1).fillna(0) + (df_corners["corner_end"] < df_corners["corner_start"].shift(-1).fillna(1000000)) * df_corners["corner_end"]

        df_corners["name"] = "T" + df_corners["number"].astype(int).astype(str)
        df_corners["type"] = "Corner"
        df_corners = df_corners[["type","name","number","X","Y","corner_start","corner_end","angle","Distance"]]


        df_straights = pd.DataFrame(
            {
                "corner_start" : df_corners.corner_end.shift(1).fillna(df_corners.corner_end.max()),
                "corner_end" : df_corners.corner_start,
                "number" : df_corners.number.shift(1).fillna(0).astype(int),
            }
        )
        df_straights["type"] = "Straight"
        df_straights["name"] = "S" + df_straights["number"].astype(str)

        self.track_regions = pd.concat([df_corners, df_straights])


class Driver:
    """
    Represents a Formula 1 driver with their associated information.

    Attributes
    ----------
    RacingNumber : str
        Driver's racing number (e.g., "1")
    BroadcastName : str
        Name used in broadcasts (e.g., "M VERSTAPPEN")
    FullName : str
        Driver's complete name (e.g., "Max VERSTAPPEN")
    Tla : str
        Three letter abbreviation (e.g., "VER")
    Line : int
        Driver's line number in timing screens
    TeamName : str
        Current team name (e.g., "Red Bull Racing")
    TeamColour : str
        Team's primary color hex code without # (e.g., "3671C6")
    FirstName : str
        Driver's first name (e.g., "Max")
    LastName : str
        Driver's last name (e.g., "Verstappen")
    Reference : str
        Driver's unique reference code (e.g., "MAXVER01")
    HeadshotUrl : str
        URL to driver's headshot image
    session : Session
        Reference to the parent session object
    """

    def __init__(
        self,
        RacingNumber: str = None,
        BroadcastName: str = None,
        FullName: str = None,
        Tla: str = None,
        Line: int = None,
        TeamName: str = None,
        TeamColour: str = None,
        FirstName: str = None,
        LastName: str = None,
        Reference: str = None,
        HeadshotUrl: str = None,
        session: "Session" = None,
        **kwargs
    ):
        self.RacingNumber = RacingNumber
        self.BroadcastName = BroadcastName
        self.FullName = FullName
        self.Tla = Tla
        self.Line = Line
        self.TeamName = TeamName
        self.TeamColour = TeamColour
        self.FirstName = FirstName
        self.LastName = LastName
        self.Reference = Reference
        self.HeadshotUrl = HeadshotUrl
        self.session = session

        # Set any additional attributes from kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)  # Removed .lower() to preserve exact key names

    def get_telemetry(self):
        """
        Get telemetry data for this driver from the session.

        Returns
        -------
        DataFrame
            Telemetry data filtered for this driver
        """
        if self.session and self.session.carTelemetry is not None:
            return self.session.carTelemetry[
                self.session.carTelemetry["DriverNo"] == self.RacingNumber
            ]
        return None

    def get_laps(self):
        """
        Get lap data for this driver from the session.

        Returns
        -------
        DataFrame
            Lap data filtered for this driver
        """
        if self.session and self.session.laps is not None:
            return self.session.laps[
                self.session.laps["DriverNo"] == self.RacingNumber
            ]
        return None

    def __repr__(self):
        return f"[{self.RacingNumber}] {self.Tla} - {self.TeamName}"


# Standard Library Imports
from urllib.parse import urljoin

# Internal Project Imports
from .livetimingf1_adapter import livetimingF1_request
from ..utils.exceptions import livef1Exception

def download_data(
    season_identifier: int = None, 
    location_identifier: str = None, 
    session_identifier: str | int = None
):
    """
    Downloads and filters F1 data based on the provided season, location, and session identifiers.

    Parameters
    ----------
    
    season_identifier : :class:`int`
        The unique identifier for the F1 season. This is a required parameter.
    location_identifier : :class:`str`
        The location (circuit or country name) for filtering meetings (races).
    session_identifier : :class:`str`
        The session name (e.g., 'FP1', 'Qualifying') or key (integer) to filter a specific session within a meeting.
    
    Returns
    ----------
    dict
        The filtered dataset containing the requested season, meeting, or session data.
    
    Raises
    ----------
    livef1Exception
        Raised if any of the required parameters are missing or if no matching data is found.

    Examples
    -------------
    .. code-block:: python
    
       print("Hello World")

    """
    
    # Initialize a variable to store the final filtered data
    last_data = None

    # Ensure a season identifier is provided (mandatory)
    if season_identifier is None:
        raise livef1Exception("Please provide at least a `season_identifier`.")

    try:
        # Download full season data using the F1 API
        season_data = livetimingF1_request(urljoin(str(season_identifier) + "/", "Index.json"))
        last_data = season_data  # Default to entire season data initially

        # If a location (race circuit) is provided, filter the season data to find the specific meeting (race)
        if location_identifier:
            meeting_data = next(
                (meeting for meeting in season_data["Meetings"] if meeting["Location"] == location_identifier), 
                None
            )
            if meeting_data:
                last_data = meeting_data  # Update with filtered meeting data
            else:
                raise livef1Exception(f"Meeting at location '{location_identifier}' not found.")
        else:
            meeting_data = season_data["Meetings"]

        # If a session (e.g., FP1, Qualifying) is provided, further filter the meeting data
        if session_identifier:
            if isinstance(session_identifier, str):
                # Filter by session name (string match)
                session_data = next(
                    (session for session in meeting_data['Sessions'] if session['Name'] == session_identifier), 
                    None
                )
            elif isinstance(session_identifier, int):
                # Filter by session key (integer match)
                session_data = next(
                    (session for session in meeting_data['Sessions'] if session['Key'] == session_identifier), 
                    None
                )
            
            if session_data:
                last_data = session_data  # Update with filtered session data
            else:
                raise livef1Exception(f"Session with identifier '{session_identifier}' not found.")

    except Exception as e:
        # Catch any exception and wrap it in a custom livef1Exception
        raise livef1Exception(e) from e

    # Return the final filtered data (season, meeting, or session)
    return last_data
